from unittest import TestCase

import numpy as np

from pdf_question_spacer.text_row_filters import RowFilter
from pdf_question_spacer.text_row_extractors import RowExtraction


class TestRowFilter(TestCase):

    def setUp(self):
        # Create test image
        image_rows = np.array([np.ones((512, 32)) for i in range(5)])
        indices = np.array([(i * 32, (i + 1) * 32) for i in range(5)])
        self.extraction = RowExtraction(image_rows, indices)

    def test_can_filter_odd_rows_from_test_image(self):
        row_filter = RowFilter(RegionPredicate())
        extraction = row_filter.filter_extraction(self.extraction)
        self.assertEqual(3, len(extraction.rows))
        self.assertEqual(3, len(extraction.row_indices))

    def test_empty_extraction_returned_on_no_matching_rows(self):
        row_filter = RowFilter(lambda region, index: False)
        extraction = row_filter.filter_extraction(self.extraction)
        self.assertEqual(0, len(extraction.rows))
        self.assertEqual(0, len(extraction.row_indices))

    def test_all_regions_returned_when_all_rows_are_matched(self):
        row_filter = RowFilter(lambda region, index: True)
        extraction = row_filter.filter_extraction(self.extraction)
        self.assertEqual(5, len(extraction.rows))
        self.assertEqual(5, len(extraction.row_indices))


class RegionPredicate():

    def __init__(self):
        self._count = -1

    def __call__(self, img, index) -> bool:
        self._count += 1
        return self._count % 2 == 0



"""
This module contains classes handling the first stage of processing the pdf:
extracting rows of text.
"""

from dataclasses import dataclass

from typing import Callable, Any

from nptyping import Array
import numpy as np


@dataclass(frozen=True)
class RowExtraction:
    """
    Encapsulates the result of a row-wise extraction from an image of
    sub-images ('regions') usually satisfying some predicate.

    This class holds the start and end indices of the regions with respect to
    the original image along with the actual regions themselves.
    """
    rows: Array[Array[Any]]
    row_indices: Array[Array[int, ..., 2]]


class TextRowExtractor:
    """
    Extract indices representing the start and end points of rows ('regions')
    matching some predicate.
    """

    def __init__(
            self,
            pixel_predicate: Callable[
                [Array[Array[Any]]], Array[Array[bool]]
            ] = lambda arr: arr == 0
    ):
        self._pixel_predicate = pixel_predicate

    @property
    def pixel_predicate(self) -> Callable[
            [Array[Array[Any]]], Array[Array[bool]]
    ]:
        return self._pixel_predicate

    @pixel_predicate.setter
    def pixel_predicate(
            self,
            pixel_predicate: Callable[[Array[Array[Any]]], Array[Array[bool]]]
    ):
        self._pixel_predicate = pixel_predicate

    def extract(self, image: Array[Array[Any]]) -> RowExtraction:
        """
        Extract 'regions' from an image matching this objects predicate.
        """
        row_contains_black = np.any(self._pixel_predicate(image), axis=1)
        runs = find_runs(1, row_contains_black)
        return RowExtraction(
            # can't apply np.fromiter to 2d arrays
            np.array([image[slice(*run)] for run in runs]),
            runs
        )


# credit:
# https://stackoverflow.com/questions/31544129/extract-separate-non-zero-blocks-from-array
def find_runs(value: Any, a: Array[Any]) -> Array[Array[int, ..., 2]]:
    """inclusive-exclusive"""
    # Create an array that is 1 where a is `value`, and pad each end with
    # an extra 0.
    isvalue = np.concatenate(([0], np.equal(a, value).view(np.int8), [0]))
    absdiff = np.abs(np.diff(isvalue))
    # Runs start and end where absdiff is 1.
    ranges = np.where(absdiff == 1)[0].reshape(-1, 2)
    return ranges



"""
This module handles the selection of image regions, the second stage of
processing.
"""

import re
from typing import Callable, Any, Sequence, Tuple

import numpy as np
import cv2
import matplotlib.pyplot as plt
import pytesseract
from nptyping import Array
from fuzzywuzzy import process

from .text_row_extractors import RowExtraction


def pad_array(
        image: Array[Array[Any]],
        pad_amount: int = 50,
        whitespace_element: Any = 255
):
    whitespace_line = np.repeat(whitespace_element, image.shape[1])
    whitespace_line = np.array(whitespace_line, dtype=np.uint8)
    return np.concatenate((
        [whitespace_line] * pad_amount,
        image,
        [whitespace_line] * pad_amount
    ))


def pad_then_extract(
        image: Array[Array[Any]],
        image_to_text_func: Callable[
            [Array[Array[Any]]], str
        ] = pytesseract.image_to_string,
        pad_amount: int = 50,
        whitespace_element: Any = 255
):
    return image_to_text_func(pad_array(image, pad_amount, whitespace_element))


class RowFilter:
    """
    Filters regions from some RowExtraction by extracting text and applying
    some predicate.
    """

    def __init__(
            self,
            region_predicate: Callable[
                [Array[Array[Any]]], bool
            ]
    ):
        self._region_predicate = region_predicate

    @property
    def region_predicate(self) -> Callable[[Array[Array[Any]]], str]:
        return self._region_predicate

    @region_predicate.setter
    def region_predicate(
            self,
            region_predicate: Callable[[Array[Array[Any]]], str]
    ):
        self._region_predicate = region_predicate

    def filter_extraction(
            self, extraction_obj: RowExtraction) -> RowExtraction:
        """
        Return a RowExtraction obj with rows taken from the passed
        RowExtraction obj, and matching the regular expression held by
        this object.
        """
        matching_indices = []
        for index, row in enumerate(extraction_obj.rows):
            if (self._region_predicate(row, index)):
                matching_indices.append(index)
        return RowExtraction(
            extraction_obj.rows[matching_indices],
            extraction_obj.row_indices[matching_indices]
        )


class TextMatcher:
    """
    Uses fuzzywuzzy to match a text from a passed image (in the form of a numpy
    array) to a string from a set of predefined strings.
    """

    def __init__(
            self,
            known_lines: Sequence[str],
            regexes: Sequence[str],
            image_to_text_func: Callable[
                [Array[Array[Any]]], str
            ] = pad_then_extract
    ):
        self._known_lines = known_lines
        self._regexes = regexes
        self._matches = []  # Store matches for convenience
        self._image_to_text_func = image_to_text_func  # For testing

    def __call__(self, row: [Array[Array[Any]]], index: int) -> bool:
        """
        Extract the text from a numpy array representing an image and test
        the best match to that string from this objects known_lines attribute
        using Levenshtein distance against this objects regexes attribute.
        """
        row_text = self._image_to_text_func(row)
        if (row_text):  # '' gives annoying fuzzywuzzy warnings
            match = process.extractOne(row_text, self._known_lines)
            self._matches.append((row_text, match[0]))
            return any(
                map(lambda regex: re.match(regex, match[0]), self._regexes)
            )
        else:
            self._matches.append((row_text, "<NO MATCH FOUND>"))
            return False

    @property
    def regexes(self) -> str:
        return self._regexes

    @regexes.setter
    def regexes(self, regexes: str):
        self._regexes = regexes

    @property
    def image_to_text_func(self) -> Callable[[Array[Array[Any]]], str]:
        return self._image_to_text_func

    @image_to_text_func.setter
    def image_to_text_func(
            self,
            image_to_text_func: Callable[[Array[Array[Any]]], str],
    ):
        self._image_to_text_func = image_to_text_func

    @property
    def known_lines(self) -> Sequence[str]:
        return self._known_lines

    @property
    def matches(self) -> Sequence[Tuple[str, str]]:
        return self._matches

    @known_lines.setter
    def known_lines(self, known_lines: Sequence[str]):
        self._known_lines = known_lines

    @classmethod
    def from_array(
            cls,
            image_as_array: Array[Array[Any]],
            regexes,
            image_to_text_func: Callable[
                [Array[Array[Any]]], str
            ] = pad_then_extract

    ) -> "TextMatcher":
        extraction = pytesseract.image_to_string(image_as_array)
        lines_of_text = str(extraction).splitlines()
        return cls(lines_of_text, regexes, image_to_text_func)


class InteractiveMatcher:
    """
    Class which allows users to select regions to add whitespace to by showing
    them the regions.
    """

    def __init__(
            self,
            all_rows: RowExtraction,
            previous_regions_preview: int = 5,
            show_previous_regions: bool = True
    ):
        self._all_rows = all_rows
        self._previous_regions_preview = previous_regions_preview
        self._show_previous_regions = show_previous_regions
        self._button_press = "n"

    def __call__(self, row: [Array[Array[Any]]], index: int) -> bool:
        """
        Show the region to the user in a matplotlib figure and let them choose
        whether to accept the region or not via a keypress.
        """
        # First of all show the preview if appropriate
        show_preview = self._show_previous_regions and index != 0
        if (show_preview):
            self.show_preview(index)

        plt.subplot(2, 1, 2 if show_preview else 1)
        plt.title("Current Region", fontweight="bold")
        plt.xlabel("Input (y/n) to add whitespace before the region")
        plt.connect('key_press_event', self.key_pressed)
        plt.imshow(row, cmap="gray")

        # Wait for keypress
        plt.waitforbuttonpress()
        plt.close()
        if (self._button_press == "y"):
            return True
        else:
            return False

    def show_preview(self, row_index: int):
        preview_length = min(self._previous_regions_preview, row_index)
        preview_slice = slice(row_index - preview_length, row_index)
        preview = cv2.vconcat([
            pad_array(region, pad_amount=20)
            for region in self._all_rows.rows[preview_slice]
        ])

        plt.subplot(2, 1, 1)
        plt.title("Previous Regions", fontweight="bold")
        plt.imshow(preview, cmap="gray")

    def key_pressed(self, event):
        self._button_press = event.key

    @property
    def previous_regions_preview(self) -> int:
        return self._previous_regions_preview

    @previous_regions_preview.setter
    def regexes(self, previous_regions_preview: int):
        self._previous_regions_preview = previous_regions_preview

    @property
    def all_rows(self) -> RowExtraction:
        return self._all_rows

    @property
    def show_previous_regions(self) -> bool:
        return self._show_previous_regions



import argparse
import re

import numpy as np
import cv2
from pdf2image import convert_from_path

from .text_row_extractors import TextRowExtractor
from .text_row_filters import RowFilter, TextMatcher, InteractiveMatcher
from .whitespace_inserter import WhitespaceInserter, ImagePager


QUESTION_REGEX = "^(\()?(([\dvi]+)|([a-z]))[\.\)]"


def parse_args():
    description = """
    Add whitespace to sections of pdfs and output the resulting images as pngs.
    """
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "infile",
        help="name of pdf to add whitespace to",
    )
    parser.add_argument(
        "whitespace_length",
        help="""Number of lines of whitespace to add per match (in pixels),
        default is 400
        """,
        type=int,
        default=400
    )

    parser.add_argument(
        "-i",
        "--interactive",
        help="""Instead of matching lines using a regular expression, show
        each region to the user and allow them to decide whether to add
        whitespace or not (overrides most other options)
        """,
        action="store_true"
    )
    parser.add_argument(
        "-pl",
        "--preview-length",
        help="""In interactive mode, for every shown region, show the previous
        <n> regions concatenated (where possible) where n is the value passed
        to this argument, default is 5. Has no effect if --interactive is not
        chosen
        """,
        type=int,
        default=5
    )

    parser.add_argument(
        "-r",
        "--regexes",
        nargs="+",
        help="""Match lines with these regular expressions, in addition
        to the default regex which matches lines which appear to be the start
        of questions""",
        default=[],
    )
    parser.add_argument(
        "-ir",
        "--ignore-default-regex",
        help="""Do not use the default regex which matches lines which appear
        to be the start of questions. To use this option you must also use
        the --regexes option""",
        action="store_true"
    )

    parser.add_argument(
        "-c",
        "--colour",
        help="Colour of the whitespace, default 255 (white)",
        default=255
    )
    parser.add_argument(
        "-s1",
        "--skip-first",
        help="Skip first regular expression match",
        action="store_true"
    )
    parser.add_argument(
        "-d",
        "--debug",
        help="""Show the text extracted from the pdf regions, along with the
        corresponging region in a matplotlib figure (not functional in
        interactive mode)""",
        action="store_true"
    )
    parser.add_argument(
        "--dpi",
        help="""The image quality, default is 400. This is passed directly
        to pdf2image.convert_from_path(), also note using large values
        will take longer and may cause crashes!""",
        type=int,
        default=400
    )
    return parser.parse_args()


def main():

    # Parse args
    args = parse_args()

    print("Opening pdf as images...")
    pil_images = convert_from_path(args.infile, dpi=args.dpi)

    print("Concatenating images...")
    numpy_images = [np.array(im_part.convert("L")) for im_part in pil_images]
    img = cv2.vconcat(numpy_images)

    print("Extracting rows of text from image...")
    extractor = TextRowExtractor()
    extraction = extractor.extract(img)

    args.regexes += [] if args.ignore_default_regex else [QUESTION_REGEX]
    if (args.interactive):
        print("Launching interactive mode...")
        matcher = InteractiveMatcher(extraction, args.preview_length)
    else:
        print("Filtering extracted rows by specified regular expression...")
        matcher = TextMatcher.from_array(img, args.regexes)
    row_filter = RowFilter(matcher)
    filtered_extraction = row_filter.filter_extraction(extraction)

    print("Inserting whitespace...")
    wspace_inserter = WhitespaceInserter(args.whitespace_length)
    img, shifted_regions = wspace_inserter.insert_whitespace(
        img,
        filtered_extraction.row_indices[args.skip_first:],
        extraction.row_indices,
        args.colour
    )

    print("Creating pdf pages...")
    pager = ImagePager(numpy_images[0].shape[0])
    pages = pager.organize_pages(
        img, shifted_regions, args.colour
    )

    for index, page in enumerate(pages):
        cv2.imwrite("out{index}.png".format(index=index), page)

    if (args.debug and not args.interactive):
        import matplotlib.pyplot as plt
        for index, row in enumerate(extraction.rows):
            plt.imshow(row, cmap="gray")
            extracted = matcher.matches[index]
            plt.title(
                """
                EXTRACTED TEXT: '{extracted_text}'
                AMENDED TEXT: '{amended_text}'
                """.format(
                    extracted_text=extracted[0], amended_text=extracted[1]
                )
            )
            plt.title(
                "MATCHES A REGEX: {does_match}".format(
                    does_match=str(any(map(
                        lambda regex: re.match(regex, extracted[1]),
                        args.regexes
                    )))
                ),
                loc="right"
            )
            plt.show()


if __name__ == "__main__":
    main()



from unittest import TestCase

import numpy as np

from pdf_question_spacer.whitespace_inserter import (
    WhitespaceInserter, ImagePager
)


class TestWhitespaceInsert(TestCase):

    def setUp(self):
        self.inserter = WhitespaceInserter(2, True)
        self.pager = ImagePager(4)

    def test_whitespace_inserted_correctly_for_one_region(self):
        arr = np.array([
            [0, 0, 0, 0],
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [0, 0, 0, 0],
        ])
        regions = np.array([
            [1, 3]  # inclusive-exclusive range
        ])
        img, _ = self.inserter.insert_whitespace(
            arr, regions, regions, 2
        )

        np.testing.assert_array_equal(
            np.array([
                [0, 0, 0, 0],
                [2, 2, 2, 2],
                [2, 2, 2, 2],
                [1, 1, 1, 1],
                [1, 1, 1, 1],
                [0, 0, 0, 0],
            ]),
            img
        )

    def test_region_indices_shifted_correctly_for_one_region(self):
        arr = np.array([
            [0, 0, 0, 0],
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [0, 0, 0, 0],
        ])
        regions = np.array([
            [1, 3]
        ])
        _, shifted_regions = self.inserter.insert_whitespace(
            arr, regions, regions, 2
        )

        np.testing.assert_array_equal(
            np.array([
                [3, 5]
            ]),
            shifted_regions
        )

    def test_no_whitespace_added_if_regions_empty(self):
        arr = np.array([
            [0, 0, 0, 0],
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [0, 0, 0, 0],
        ])
        regions = np.array([])
        img, _ = self.inserter.insert_whitespace(
            arr, regions, regions, 2
        )
        np.testing.assert_array_equal(
            arr, img
        )

    def test_whitespace_inserted_correctly_for_one_region_at_array_end(self):
        arr = np.array([
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [1, 1, 1, 1],
            [1, 1, 1, 1],
        ])
        regions = np.array([
            [2, 4]
        ])
        img, _ = self.inserter.insert_whitespace(
            arr, regions, regions, 2
        )

        np.testing.assert_array_equal(
            np.array([
                [0, 0, 0, 0],
                [0, 0, 0, 0],
                [2, 2, 2, 2],
                [2, 2, 2, 2],
                [1, 1, 1, 1],
                [1, 1, 1, 1],
            ]),
            img
        )

    def test_whitespace_inserted_correctly_for_one_region_at_array_start(self):
        arr = np.array([
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
        ])
        regions = np.array([
            [0, 2]
        ])
        img, _ = self.inserter.insert_whitespace(
            arr, regions, regions, 2
        )

        np.testing.assert_array_equal(
            np.array([
                [2, 2, 2, 2],
                [2, 2, 2, 2],
                [1, 1, 1, 1],
                [1, 1, 1, 1],
                [0, 0, 0, 0],
                [0, 0, 0, 0],
            ]),
            img
        )

    def test_whitespace_inserted_correctly_for_multiple_regions(self):
        arr = np.array([
            [0, 0, 0, 0],
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [1, 1, 1, 1],     
        ])
        target_regions = np.array([
            [1, 3], [5, 6]
        ])
        all_regions = np.array([
            [1, 3], [5, 6]
        ])
        img, _ = self.inserter.insert_whitespace(
            arr, target_regions, all_regions, 2
        )
        np.testing.assert_array_equal(
            np.array([
                [0, 0, 0, 0],
                [2, 2, 2, 2],
                [2, 2, 2, 2],
                [1, 1, 1, 1],
                [1, 1, 1, 1],
                [0, 0, 0, 0],
                [0, 0, 0, 0],
                [2, 2, 2, 2],
                [2, 2, 2, 2],
                [1, 1, 1, 1],                
            ]),
            img
        )

    def test_region_indices_shifted_correctly_for_multiple_regions(self):
        arr = np.array([
            [0, 0, 0, 0],
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [1, 1, 1, 1]
        ])
        regions = np.array([
            [1, 3], [5, 6]
        ])
        _, shifted_regions = self.inserter.insert_whitespace(
            arr, regions, regions, 2
        )
        np.testing.assert_array_equal(
            np.array([
                [3, 5], [9, 10]
            ]),
            shifted_regions
        )

    def test_region_indices_shifted_correctly_with_non_target_indices(self):
        arr = np.array([
            [0, 0, 0, 0],
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
        ])
        all_regions = np.array([
            [1, 3], [4, 5]
        ])
        regions = np.array([
            [1, 3]
        ])
        _, shifted_regions = self.inserter.insert_whitespace(
            arr, regions, all_regions, 2
        )
        np.testing.assert_array_equal(
            np.array([
                [3, 5], [6, 7]
            ]),
            shifted_regions
        )

    def test_region_indices_shifted_correctly_with_non_target_index_before_target_index(self):
        arr = np.array([
            [0, 0, 0, 0],
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
        ])
        all_regions = np.array([
            [0, 1], [1, 3]
        ])
        regions = np.array([
            [1, 3]
        ])
        _, shifted_regions = self.inserter.insert_whitespace(
            arr, regions, all_regions, 2
        )
        np.testing.assert_array_equal(
            np.array([
                [0, 1], [3, 5]
            ]),
            shifted_regions
        )

    def test_no_whitespace_added_if_regions_dont_span_pages(self):
        arr = np.array([
            [0, 0, 0, 0],
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [0, 0, 0, 0],
            # Page 2
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
        ])
        regions = np.array([
            [1, 3], [5, 6]
        ])
        pages = self.pager.organize_pages(
            arr, regions, 2
        )

        np.testing.assert_array_equal(
            np.array([
                [[0, 0, 0, 0],
                 [1, 1, 1, 1],
                 [1, 1, 1, 1],
                 [0, 0, 0, 0]],
                # Page 2
                [[0, 0, 0, 0],
                 [0, 0, 0, 0],
                 [0, 0, 0, 0],
                 [0, 0, 0, 0]]
            ]),
            pages
        )

    def test_whitespace_added_one_region_spans_page(self):
        arr = np.array([
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [1, 1, 1, 1],
            # Page 2
            [1, 1, 1, 1],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
        ])
        regions = np.array([
            [1, 3], [3, 5]
        ])
        pages = self.pager.organize_pages(
            arr, regions, 2
        )

        np.testing.assert_array_equal(
            np.array([
                [[0, 0, 0, 0],
                 [0, 0, 0, 0],
                 [0, 0, 0, 0],
                 [2, 2, 2, 2]],
                # Page 2
                [[1, 1, 1, 1],
                 [1, 1, 1, 1],
                 [0, 0, 0, 0],
                 [0, 0, 0, 0]],
                # Page 3
                [[0, 0, 0, 0],
                 [2, 2, 2, 2],
                 [2, 2, 2, 2],
                 [2, 2, 2, 2]]
            ]),
            pages
        )

    def test_whitespace_added_multiple_regions_span_page(self):
        arr = np.array([
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [1, 1, 1, 1],
            # Page 2
            [1, 1, 1, 1],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            # Page 3
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
        ])
        regions = np.array([
            [1, 2], [3, 5], [6, 9]
        ])
        pages = self.pager.organize_pages(
            arr, regions, 2
        )

        np.testing.assert_array_equal(
            np.array([
                [[0, 0, 0, 0],
                 [0, 0, 0, 0],
                 [0, 0, 0, 0],
                 [2, 2, 2, 2]],
                # Page 2
                [[1, 1, 1, 1],
                 [1, 1, 1, 1],
                 [0, 0, 0, 0],
                 [2, 2, 2, 2]],
                # Page 3
                [[0, 0, 0, 0],
                 [0, 0, 0, 0],
                 [0, 0, 0, 0],
                 [0, 0, 0, 0]],
                # Page 4
                [[0, 0, 0, 0],
                 [0, 0, 0, 0],
                 [2, 2, 2, 2],
                 [2, 2, 2, 2]],
            ]),
            pages
        )



"""
This module contains classes used for inserting whitespace into the appropriate
places in the pdf, the last stage of processing.
"""
from typing import Any, Tuple

import numpy as np
from nptyping import Array


class WhitespaceInserter:
    """Inserts whitespace into numpy array images."""

    def __init__(self, whitespace_length: int, add_before_region: bool = True):
        self.whitespace_length = whitespace_length
        self.add_before_region = add_before_region

    def insert_whitespace(
            self,
            img: Array[Array[Any]],
            target_region_indices: Array[Array[int, ..., 2]],
            all_region_indices: Array[Array[int, ..., 2]],
            whitespace_element: Any
    ) -> Tuple[Array[Array[Any]], Array[Array[int, ..., 2]]]:
        """
        Insert whitespace into an image (a 2d numpy array) inbetween the
        regions of the image specified by 'target_region_indices'. Return
        the image with whitespace added along with the shifted regions.
        """
        whitespace_line = np.repeat(whitespace_element, img.shape[1])
        for index, (start, end) in enumerate(target_region_indices):
            if (self.add_before_region):
                slice_point = start + index * self.whitespace_length
            else:
                slice_point = end + index * self.whitespace_length + 1
            img = pad_array(
                img, slice_point, whitespace_line, self.whitespace_length
            )
            all_region_indices = add_to_regions_above(
                slice_point, all_region_indices, self.whitespace_length
            )
        return (img, all_region_indices)


class ImagePager:
    """
    Breaks image arrays into multiple arrays ('pages') in such a way that
    no text is cut in half.
    """

    def __init__(self, page_pixel_length: int):
        self.page_pixel_length = page_pixel_length

    def pad_regions_spanning_pages(
            self,
            img: Array[Array[Any]],
            regions: Array[Array[int, ..., 2]],
            whitespace_element: Any
    ) -> Array[Array[Array[Any]]]:
        """
        Add lines of whitespace before regions that would otherwise span
        multiple pages. The new padded image array is then returned.
        """
        region_pages = regions // self.page_pixel_length
        overflowing_regions = regions[region_pages[:, 0] != region_pages[:, 1]]
        whitespace_line = np.repeat(whitespace_element, img.shape[1])
        if (overflowing_regions.size != 0):
            region_start = overflowing_regions[0, 0]
            pad_amount = inv_mod(region_start, self.page_pixel_length)
            img = pad_array(img, region_start, whitespace_line, pad_amount)
            regions = add_to_regions_above(region_start, regions, pad_amount)
            return self.pad_regions_spanning_pages(
                img, regions, whitespace_element
            )
        else:
            return img

    def split_into_pages(
            self,
            img: Array[Array[Any]],
            regions: Array[Array[int, ..., 2]],
            whitespace_element: Any
    ) -> Array[Array[Array[Any]]]:
        """
        Return an array of image arrays, with padding added to the last array
        so that all arrays are of the same shape.
        """
        end_padding = inv_mod(img.shape[0], self.page_pixel_length)
        if (end_padding):
            whitespace_line = np.repeat(whitespace_element, img.shape[1])
            img = np.concatenate((img, [whitespace_line] * end_padding))
        return np.array(np.array_split(
            img, img.shape[0] / self.page_pixel_length
        ))

    def organize_pages(
            self,
            img: Array[Array[Any]],
            regions: Array[Array[int, ..., 2]],
            whitespace_element: Any
    ) -> Array[Array[Array[Any]]]:
        """
        Returns an array of images (as arrays of size corresponding to
        page_pixel_length) such that none of the regions of the image specified
        by the parameter 'regions' span more than one page.
        Whitespace corresponding to the 'whitespace' parameter is added as
        appropriate.

        Is a wrapper around pad_regions_spanning_pages() and split_into_pages()
        """
        shifted_img = self.pad_regions_spanning_pages(
            img, regions, whitespace_element
        )
        return self.split_into_pages(
            shifted_img, regions, whitespace_element
        )


def inv_mod(dividend: int, divisor: int):
    """
    Return the additive inverse of 'dividend' wrt to modulo divisor arithmetic,
    0 <= return < divisor.
    """
    return (divisor - (dividend % divisor)) % divisor


def pad_array(
        array: Array[Any],
        index: int,
        pad_element: Any,
        amount: int
) -> Array[Any]:
    """
    Add 'amount' sets of 'pad_element' to the specified array at the index
    specified by 'index'.
    """
    return np.concatenate((
        array[:index],
        [pad_element] * amount,
        array[index:]
    ))


def add_to_regions_above(
        cut_off: int,
        regions: Array[Array[int, ..., 2]],
        amount: int,
        include_equality: bool = True
):
    if (include_equality):
        shifted_arr = (regions[:, 0] >= cut_off) * amount
    else:
        shifted_arr = (regions[:, 0] > cut_off) * amount
    return regions + shifted_arr.reshape(len(regions), 1)



from unittest import TestCase

import numpy as np

from pdf_question_spacer.text_row_extractors import TextRowExtractor


class TestRowExtractor(TestCase):

    def setUp(self):
        pass

    def test_can_extract_one_row(self):
        arr = np.array([
            [1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1],
            [0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0],
            [1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1]
        ], dtype=np.bool)
        extractor = TextRowExtractor()
        extracted_regions = extractor.extract(arr)
        # Assert region correct
        np.testing.assert_array_equal(
            np.array(
                [[0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0]],
                dtype=np.bool),
            extracted_regions.rows[0]
        )
        self.assertEqual(1, len(extracted_regions.rows))

    def test_can_extract_one_row_at_start(self):
        arr = np.array([
            [0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0],
            [1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1]
        ], dtype=np.bool)
        extractor = TextRowExtractor()
        extracted_regions = extractor.extract(arr)
        # Assert region correct
        np.testing.assert_array_equal(
            np.array(
                [[0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0]],
                dtype=np.bool),
            extracted_regions.rows[0]
        )
        # Assert no more regions found
        self.assertEqual(1, len(extracted_regions.rows))

    def test_can_extract_one_row_at_end(self):
        arr = np.array([
            [1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1],
            [0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0]
        ], dtype=np.bool)
        extractor = TextRowExtractor()
        extracted_regions = extractor.extract(arr)
        # Assert region correct
        np.testing.assert_array_equal(
            np.array(
                [[0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0]],
                dtype=np.bool),
            extracted_regions.rows[0]
        )
        self.assertEqual(1, len(extracted_regions.rows))

    def test_can_extract_multiple_regions(self):
        arr = np.array([
            [1, 1, 1, 1, 1, 1],
            [0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0],
            [1, 1, 1, 1, 1, 1],
            [0, 0, 0, 0, 0, 0],
            [1, 1, 1, 1, 1, 1]
        ], dtype=np.bool)
        extractor = TextRowExtractor()
        extracted_regions = extractor.extract(arr)
        # Assert region correct
        np.testing.assert_array_equal(
            np.array(
                [[0, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0, 0]],
                dtype=np.bool),
            extracted_regions.rows[0]
        )
        np.testing.assert_array_equal(
            np.array(
                [[0, 0, 0, 0, 0, 0]],
                dtype=np.bool),
            extracted_regions.rows[1]
        )
        # Assert two regions found in total
        self.assertEqual(2, len(extracted_regions.rows))

    def test_can_extract_region_not_all_target(self):
        arr = np.array([
            [1, 1, 1, 1, 1, 1],
            [1, 0, 1, 1, 0, 0],
            [1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1]
        ], dtype=np.bool)
        extractor = TextRowExtractor()
        extracted_regions = extractor.extract(arr)
        # Assert region correct
        np.testing.assert_array_equal(
            np.array(
                [[1, 0, 1, 1, 0, 0]],
                dtype=np.bool),
            extracted_regions.rows[0]
        )
        self.assertEqual(1, len(extracted_regions.rows))

    def test_extract_returns_nothing_on_no_valid_rows(self):
        arr = np.array([
            [1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1]
        ], dtype=np.bool)
        extractor = TextRowExtractor()
        extracted_regions = extractor.extract(arr)
        self.assertEqual(0, len(extracted_regions.rows))



from setuptools import setup, find_packages


setup(
    name="pdf-question-spacer",
    packages=find_packages(),
    test_suite="tests",
    include_package_data=True,
    python_requires='>=3.5',
    url="https://github.com/LaurenceWarne/pdf-question-spacer",
    entry_points={
        "console_scripts":
        [
            "space-pdf=pdf_question_spacer.space_pdf:main",
        ],
    },
    version="1.0",
    author="Laurence Warne",
    license="MIT",
    install_requires=[
        "numpy", "opencv-python", "pytesseract", "nptyping",
        "fuzzywuzzy", "matplotlib", "python-Levenshtein",
        "pdf2image", "Pillow"
    ],
    zip_safe=False
)
